from epic.config.genomes import get_effective_genome_length
from scipy.stats import poisson, rankdata


def compute_fdr(df, total_chip_reads, total_input_reads, args):

    total_island_input_reads = df["Input"].sum()

    # Hack needed in case we run on test data
    if total_island_input_reads == 0:
        total_island_input_reads = 2

    scaling_factor = (total_chip_reads * 1.0) / total_input_reads

    effective_genome_size = get_effective_genome_length(args.genome)
    zero_controls_multiplier = total_input_reads * 1.0 / effective_genome_size

    avg_0_denom = (df["End"] - df["Start"] + 1) * zero_controls_multiplier
    avg_0_denom[avg_0_denom > 0.25] = 0.25
    avg_0_denom = avg_0_denom * scaling_factor

    avg = df["Input"] * scaling_factor
    avg[df["Input"] == 0] = avg_0_denom[df["Input"] == 0]

    df["P_value"] = poisson.sf(df["ChIP"], avg)
    no_differential_expression = df["ChIP"] <= avg
    df.loc[no_differential_expression, "P_value"] = 1

    df["Fold_change"] = df["ChIP"] / avg

    ranked_p_values = rankdata(df["P_value"])
    df["FDR_value"] = df["P_value"] * len(df) / ranked_p_values
    fdr_too_high = df["FDR_value"] > 1
    df.loc[fdr_too_high, "FDR_value"] = 1

    df = df[df["FDR_value"] < args.false_discovery_rate_cutoff]

    return df



import pandas as pd


def merge_chromosome_dfs(df_tuple):

    plus_df, minus_df = df_tuple
    if plus_df.empty:
        return minus_df
    if minus_df.empty:
        return plus_df

    df = pd.concat([plus_df, minus_df]).sort_values(by="Bin").reset_index(
        drop=True)

    df["Count"] = df.groupby(["Bin"])["Count"].transform("sum")

    # Only need one entry per bin
    df = df.drop_duplicates(["Chromosome", "Bin"])

    df[["Bin", "Count"]] = df[["Bin", "Count"]]

    return df.reset_index(drop=True)



from sys import stdout
from itertools import chain
from collections import OrderedDict
import logging

import pandas as pd

from natsort import natsorted

from epic.windows.count.count_reads_in_windows import count_reads_in_windows
from epic.config.genomes import create_genome_size_dict
from epic.statistics.compute_background_probabilites import compute_background_probabilities
from epic.statistics.count_to_pvalue import count_to_pvalue
from epic.statistics.fdr import compute_fdr
from epic.utils.helper_functions import merge_chip_and_input, get_total_number_of_reads, merge_same_files
from epic.windows.cluster.find_islands import find_islands


def run_epic(args):

    chip_windows = multiple_files_count_reads_in_windows(args.treatment, args)
    input_windows = multiple_files_count_reads_in_windows(args.control, args)

    chip_merged = _merge_files(chip_windows.values(), args.number_cores)
    input_merged = _merge_files(input_windows.values(), args.number_cores)

    nb_chip_reads = get_total_number_of_reads(chip_merged)
    nb_input_reads = get_total_number_of_reads(input_merged)

    merged_dfs = merge_chip_and_input(chip_merged, input_merged,
                                      args.number_cores)

    score_threshold, island_enriched_threshold, average_window_readcount = \
        compute_background_probabilities(nb_chip_reads, args)

    dfs = count_to_pvalue(merged_dfs, island_enriched_threshold,
                          average_window_readcount, args.number_cores)

    dfs = find_islands(dfs, score_threshold, args)

    logging.info("Done finding islands.")
    logging.info("Concating dfs.")
    df = pd.concat([df for df in dfs if not df.empty])
    logging.info("Labeling island bins.")

    logging.info("Computing FDR.")
    df = compute_fdr(df, nb_chip_reads, nb_input_reads, args)

    df[["Start", "End"]] = df[["Start", "End"]].astype(int)
    df.to_csv(stdout, index=False, sep=" ")


def get_island_bins(df, window_size, genome):
    """Finds the enriched bins in a df."""

    # need these chromos because the df might not have islands in all chromos
    chromosomes = natsorted(list(create_genome_size_dict(genome)))

    chromosome_island_bins = {}
    df_copy = df.reset_index(drop=False)
    for chromosome in chromosomes:
        cdf = df_copy.loc[df_copy.Chromosome == chromosome]
        if cdf.empty:
            chromosome_island_bins[chromosome] = set()
        else:
            island_starts_ends = zip(cdf.Start.values.tolist(),
                                     cdf.End.values.tolist())
            island_bins = chain(*[range(
                int(start), int(end), window_size)
                                  for start, end in island_starts_ends])
            chromosome_island_bins[chromosome] = set(island_bins)

    return chromosome_island_bins


def multiple_files_count_reads_in_windows(bed_files, args):
    """Use count_reads on multiple files and store result in dict.

    Untested since does the same thing as count reads."""

    bed_windows = OrderedDict()
    for bed_file in bed_files:
        logging.info("Binning " + bed_file)
        chromosome_dfs = count_reads_in_windows(bed_file, args)
        bed_windows[bed_file] = chromosome_dfs

    return bed_windows


def _merge_files(windows, nb_cpu):
    """Merge lists of chromosome bin df chromosome-wise.

    windows is an OrderedDict where the keys are files, the values are lists of
    dfs, one per chromosome.

    Returns a list of dataframes, one per chromosome, with the collective count
    per bin for all files."""

    windows = iter(windows)
    merged = next(windows)

    for chromosome_dfs in windows:
        merged = merge_same_files(merged, chromosome_dfs, nb_cpu)

    return merged



import pytest

from io import StringIO
import pandas as pd
from numpy import allclose, array_equal, int32

from epic.windows.count.count_reads_in_windows import count_reads_in_windows, _count_reads_in_windows_paired_end

import argparse
parser = argparse.ArgumentParser()

parser.add_argument(
    '--number-cores',
    '-cpu',
    required=False,
    default=1,
    type=int,
    help=
    '''Number of cpus to use. Can use at most one per chromosome. Default: 1.''')

parser.add_argument('--genome',
                    '-gn',
                    required=False,
                    default="hg19",
                    type=str,
                    help='''Which genome to analyze. Default: hg19.''')

parser.add_argument(
    '--keep-duplicates',
    '-k',
    required=False,
    default=False,
    type=bool,
    help=
    '''Keep reads mapping to the same position on the same strand within a library. Default is to remove all but the first duplicate.
                   ''')

parser.add_argument(
    '--window-size',
    '-w',
    required=False,
    default=200,
    type=int,
    help=
    '''Size of the windows to scan the genome. WINDOW_SIZE is the smallest possible island. Default 200.
                   ''')

parser.add_argument(
    '--fragment-size',
    '-fs',
    required=False,
    default=150,
    type=int,
    help=
    '''Size of the sequenced fragment. The center of the the fragment will be taken as half the fragment size. Default 150.
                   ''')

parser.add_argument('--paired-end',
                    '-pe',
                    required=False,
                    default=False,
                    type=bool,
                    help='''(Bam-only) Use paired-end reads.''')

@pytest.mark.integration
def test_count_reads_in_windows(expected_result, input_bed_file):

    args = parser.parse_args(['-fs', '150'])
    args.genome, args.fragment_size, args.window_size = "hg19", 150, 200
    args.keep_duplicates = False
    args.number_cores = 1

    actual_result = count_reads_in_windows(input_bed_file, args)

    actual_result = pd.concat(actual_result)

    print(actual_result)
    print(expected_result)

    assert array_equal(actual_result, expected_result)


@pytest.fixture
def expected_result():

    df = pd.read_table(
        StringIO(u"""Count Chromosome Bin
    2 chr1 39036800
    1 chr1 73781000
    1 chr1 90059800
    1 chr3 55648200
    1 chr7 20246600
    1 chr7 91135000
    1 chr13 100938400
    1 chr19 43528800
    1 chr19 47108800"""),
        sep=r"\s+",
        dtype={"Count": int32,
               "Bin": int32})
    return df


@pytest.fixture
def input_bed_file(tmpdir):
    bed_file = tmpdir.join("read_file.bed")
    bed_file.write("""chr1	39036822	39036847	U0	0	+
chr1	39036870	39036900	U0	0	-
chr1	73781101	73781126	U0	0	+
chr1	90059861	90059886	U0	0	-
chr3	55648137	55648162	U0	0	+
chr7	20246668	20246693	U0	0	+
chr7	91135110	91135135	U0	0	+
chr13	100938475	100938500	U0	0	+
chr13	115816130	115816155	U0	0	+
chr19	43528773	43528798	U0	0	+
chr19	47109000	47109025	U0	0	-""")
    return str(bed_file)


@pytest.fixture
def paired_end(tmpdir):

    bed_file = tmpdir.join("paired_end_file.bed")
    bed_file.write(
        """chr1	33076	33165	chr1	33076	33165	2PJ3LS1:183:C5RR7ACXX:7:1307:11171:75017	0	+	-
chr1	33076	33165	chr1	33076	33165	2PJ3LS1:183:C5RR7ACXX:8:1115:18054:85095	0	+	-
chr1	33660	33744	chr1	33660	33744	2PJ3LS1:183:C5RR7ACXX:2:2201:20777:74838	0	+	-
chr1	33660	33744	chr1	33660	33744	2PJ3LS1:183:C5RR7ACXX:3:1213:9620:83082	0	+	-
chr1	38474	38571	chr1	38474	38571	2PJ3LS1:183:C5RR7ACXX:4:1216:1576:83235	0	+	-
chr1	41985	42043	chr1	42040	42131	2PJ3LS1:183:C5RR7ACXX:3:1214:11778:16074	1	+	-
chr1	42223	42324	chr1	42285	42386	2PJ3LS1:183:C5RR7ACXX:3:1215:18097:73129	0	+	-""")
    return str(bed_file)


# @pytest.mark.integration
def test_count_reads_in_windows_paired_end(paired_end):

    genome, fragment_size, window_size = "hg19", 150, 200
    keep_duplicates = False

    # result = _count_reads_in_windows_paired_end(paired_end, genome,
    #                                            fragment_size, window_size,
    #                                            keep_duplicates, "chr1")

    # print(result)
    "Paired end support not implemented yet."
    # assert 0
    # assert result == expected_result
